'use strict';

/**
 * Client-side faceted filters.
 *
 * @param {Element} targetElement
 * @param {array} settings
 */
function Csff(targetElement, settings){

  /**
   * @var {object}
   */
  const _this = this;

  /**
   * @var {object}
   */
  const defaultSettings = {
    // General settings:
    injection_selector: '#facets', // Facets container
    injection_method: 'beforeend', // Facets container injection method

    // Facet filters:
    facets_show: true, // Enable faceted filters
    facets: [
      // EXAMPLE:
      // {
      //   id: 'color',
      //   title: 'Color',
      //   description: 'Filter by color:',
      //   count_show: true,
      //   values_sort: 'value',
      //   values_sort_direction: 'asc',
      //   multivalue_separator: '|',
      //   weight: 0,
      // },
      // {
      //   id: 'size',
      //   title: 'Size',
      //   description: 'Filter by size:',
      //   count_show: true,
      //   values_sort: 'value',
      //   values_sort_direction: 'asc',
      //   multivalue_separator: '|',
      //   weight: 0,
      // }
    ],
    facets_values_count_show: true, // Show values count
    facets_values_sort: 'value', // Sort values by
    facets_values_sort_direction: 'asc', // Sort direction
    facets_multivalue_separator: '|', // Split multiple values by

    // Fulltext search:
    fulltext_search_show: false, // Enable full text search
    fulltext_search_title: 'Search',
    fulltext_search_description: '',
    fulltext_search_placeholder: '',
    fulltext_search_delay: 500,

    // Reset button:
    reset_button_show: true, // Show reset button
    reset_button_title: 'Reset',
    reset_button_description: '',
    reset_button_label: 'Reset', // Reset button label

    // Empty text:
    results_empty_text: 'No matching results.',

    // HTML templates:
    template_facets_container: '<section class="csff-facets-container">\
        <form class="csff-facets-container__form" action="#"></form>\
      </section>',
    template_facet: '<section class="csff-facet view-faceted-filters-js-facet block block--dynamic" id="csff-facet-${id}">\
        <h2 class="csff-facet__title block-title block__title">${title}</h2>\
        <div class="csff-facet__description block__description">${description}</div>\
        <div class="csff-facet__content block__content content">${content}</div>\
      </section>',
    template_facet_value: '<div class="csff-facet-value form-item form-type-checkbox"><input type="checkbox" name="${facet_id}" id="${id}" value="${value}" class="csff-facet-value__value form-checkbox" /><label for="${id}" class="option csff-facet-value__label">${value} <span class="csff-facet-value__count">(${count})</span></label></div>',
    template_facet_reset: '<div class="csff-reset">\
          <button class="button button--reset">${label}</button>\
        </div>',
    template_search: '<div class="csff-search"><input type="text" name="csff-search" class="csff-search__value" placeholder="${placeholder}" /></div>',
    template_results_empty: '<div class="csff-results-empty hidden--results-empty"><div class="csff-results-empty__content content">${content}</div></div>',

    // The following options are typically not changed, but we even allow that
    // for resolving possible conflicts:
    _csff_controls_facets_container_selector: '.csff-facets-container',
    _csff_item_selector: '.csff-item',
    _csff_subitem_selector: '.csff-subitem',
    _csff_facet_class_prefix: 'csff-facet-',
    _csff_facet_data_attribute_prefix: 'csff-facet-',
    _csff_facet_data_attribute_value_suffix: '--value',
    _csff_search_class: 'csff-search',
    _csff_search_data_attribute_prefix: 'csff-search--value',
  }
  settings = {...defaultSettings, ...settings};

  /**
   * @var array
   */
  const facetIndex = {
    // EXAMPLE:
    // color: { // The facet id
    //   values: { // The items with values for the facet
    //     // Indexed by value:
    //     'Red': {
    //       value: 'Red',
    //       items: [] // May have several items with that value
    //     }
    //   }
    // }
  };

  /**
   * Initialize
   */
  this.init = function() {
    // Initialize controls:
    _this._initControls();
  }

  /**
   * Helper function to get the representative (parent) DOM Element
   * for the given elem.
   *
   * This is required to find the element to hide based on the given
   * (e.g. child) element which contains the facet properties.
   *
   * @param {Element} elem
   * @return {Element}
   */
  this._getRepresentativeItem = function (elem) {
    return elem.closest(settings._csff_item_selector);
  }

  /**
   * Internal helper function to Initialize all controls.
   *
   * @return {void}
   */
  this._initControls = function() {
    // Insert wrapping container:
    const injection_target = _this._domGetInjectionTargetElement();
    if (!injection_target) {
      throw "Injection target could not be found."
    }
    injection_target.insertAdjacentHTML(settings.injection_method, settings.template_facets_container)

    // Set the elemFacetsContainer variable to store the container:
    const elemFacetsContainer = _this._domGetFacetsContainerElement();
    if (!elemFacetsContainer) {
      throw "Facets container could not be found.";
    }

    if (settings.facets_show) {
      _this._initControlsFacets();
    }
    if(settings.fulltext_search_show){
      _this._initControlsSearch();
    }
    if(settings.reset_button_show){
      _this._initControlsReset();
    }

    // Insert hidden empty results container:
    if (settings.results_empty_text) {
      targetElement.insertAdjacentHTML('beforeend', _this._theme_results_empty(settings.results_empty_text));
    }

  }

  /**
   * Internal helper function to Initialize the facet controls.
   *
   * @return {void}
   */
  this._initControlsFacets = function() {
    // Collect the properties to build the facets:
    _this._collectProperties();

    const facets = settings.facets;
    let facetsHtml = '';
    // Sort facets by weight:
    facets.sort((a,b) => a.weight - b.weight);
    settings.facets.forEach((facet, index) => {
      if (facet.id) {
        const facet_id = facet.id;

        // Check if there are values for this facet id:
        if (facetIndex.hasOwnProperty(facet_id)) {
          let facetValuesHTML = '';
          // TODO: Sort facet values!
          // Loop values and add them as checkboxes:
          for (const [key, properties] of Object.entries(facetIndex[facet_id]['values'])) {
            const value = properties.value;
            const facetValueElementHTML = _this._theme_facet_value(
              facet_id,
              value,
              properties.items.length // TODO: This might have to be calculated only for visible elements!
            )
            facetValuesHTML += facetValueElementHTML;
          }

          // Values for this facet are present:
          // Create a container for this facet:
          facetsHtml += _this._theme_facet(
            facet_id,
            facet.title,
            facet.description,
            facetValuesHTML
          );
        }
      }
    })
    if (facetsHtml.length > 0) {
      _this._domGetFacetsContainerElement().insertAdjacentHTML('beforeend', facetsHtml)
    }

    // As all HTML was dynamically added, we have to bind the change event afterwards:
    _this._domGetFacetFilterInputElements().forEach((input, index) => {
      input.addEventListener('change', function(e) {
        _this.filter(e.target);
      })
    })
  }

  this._facetIndexGetFacetValues = function (facet_id) {
    if (facetIndex.hasOwnProperty(facet_id)) {
      return facetIndex[facet_id]['values'];
    } else {
      return {};
    }
  }

  this._facetIndexFilterFacet = function (facet_id, filterValues = []) {
    const values = _this._facetIndexGetFacetValues(facet_id);
    let results = [];
    filterValues.forEach((filterValue, index) => {
      if (values.hasOwnProperty(filterValue)) {
        results.push(values[filterValue]);
      }
    })
    return results;
  }

  /**
   * Helper function to collect all facet related properties (filterable values)
   * from the items.
   *
   * Adds all collected properties on this.facetIndex.
   *
   * @return {void}
   */
  this._collectProperties = function() {
    const items = _this._domGetItemElements();
    items.forEach((item, index) => {
      settings.facets.forEach((facet, index) => {
        if (facet.id) {
          let facetId = facet.id;
          // Find defintion of facets data attribute:
          let dataAttributeValueName = 'data-' + settings._csff_facet_data_attribute_prefix + facetId + settings._csff_facet_data_attribute_value_suffix;
          let elem = false;
          if (item.hasAttribute(dataAttributeValueName)) {
            // The element defines the data attribute itself:
            elem = item;
          } else {
            // Get child element with data attribute:
            elem = item.querySelector('[' + dataAttributeValueName + ']');
          }

          if (elem) {
            let value = elem.getAttribute(dataAttributeValueName);
            // Split value by separator as separate values:
            const values = value.split(facet.multivalue_separator)
            values.forEach((splitValue) => _this._addToIndex(facetId, splitValue, elem))

            // TODO: remove DEBUG output:
            elem.className += ' csff-item-' + facetId +'--indexed';
          }
        } else {
          console.warn('Missing "id" for facet definition: ' + JSON.stringify(facet))
        }
      })
    });
  }

  /**
   * Helper function to add facet properties to the index on this.facetIndex.
   *
   * @param {string} facetId
   * @param {string} value
   * @param {Element} elem
   *
   * @return {void}
   */
  this._addToIndex = function (facetId, value, elem) {
    // Trim value:
    value = value.trim()
    if (!facetIndex.hasOwnProperty(facetId)) {
      // No property for the facetId exists yet:
      facetIndex[facetId] = {
        values: {}
      };
    }

    if (!facetIndex[facetId]['values'].hasOwnProperty(value)) {
      // No such value exists yet for the given facetId:
      facetIndex[facetId]['values'][value] = {
        'value': value,
        'items': [],
      }
    }

    // Add the item element to the index:
    facetIndex[facetId]['values'][value]['items'].push(elem);
  }

  /**
   * Internal helper function to Initialize the full text search controls.
   */
  this._initControlsSearch = function() {
    _this._domGetFacetsContainerElement().insertAdjacentHTML('afterbegin', _this._theme_search(settings.fulltext_search_title, settings.fulltext_search_description, settings.fulltext_search_placeholder));

    let typeout;
    _this._domGetSearchInputElement().addEventListener('keyup', (e) => {
      if (typeout){
        clearTimeout(typeout);
      }
      typeout = setTimeout(() => {
        _this.filter(e.target)
      }, settings.fulltext_search_delay);
    });
  }

  this._initControlsReset = function () {
    _this._domGetFacetsContainerElement().insertAdjacentHTML('beforeend', _this._theme_facet_reset(settings.reset_button_title, settings.reset_button_description, settings.reset_button_label))
    _this._domGetResetButtonInputElement().addEventListener('click', function(e) {
      _this.reset();
    })
  }

  this.filter = function(triggeringElement) {
    // Filter by search value
    _this._filterBySearch(_this._domGetSearchInputElement().value)

    let activeFacetFilters = {
      // field_tags: ['Tag1', 'Tag2'],
      // // AND
      // field_title: ['Test']
    };
    _this._domGetFacetFilterInputElements().forEach((item, index) => {
      if(item.checked){
        activeFacetFilters[item.name] = activeFacetFilters[item.name] || []
        activeFacetFilters[item.name].push(item.value)
      }
    })
    _this._domGetItemElements().forEach((item, index) => {
      if (Object.entries(activeFacetFilters).length > 0) {
        // There are active filters, initially hide all:
        item.classList.add('hidden--filter');
      } else {
        // There are no active filters, unhide all:
        item.classList.remove('hidden--filter');
      }
    })

    // Apply the filter logic:
    let filterResultItems = [];
    for (const [facet, values] of Object.entries(activeFacetFilters)) {
      // Get the results from the single facet. Values of a single facet
      // are always combined by OR (UNION)
      const facetFilterResults = _this._facetIndexFilterFacet(facet, values);
      const facetFilterResultsItems = new Set();
      if (Object.entries(facetFilterResults).length > 0) {
        facetFilterResults.forEach((facetFilterResult) => {
          facetFilterResult.items.forEach((facetFilterResultItem) => facetFilterResultsItems.add(facetFilterResultItem))
        })

        // Now combine them with other Facets.
        // This is always done by AND (INTERSECT):
        if (filterResultItems.length === 0) {
          // No other results yet. We can't intersect with empty.
          filterResultItems = [...facetFilterResultsItems];
        } else {
          // Intersect with other facet results
          filterResultItems = filterResultItems.filter((filterResult) => [...facetFilterResultsItems].includes(filterResult))
        }
      }
    }
    if (filterResultItems.length > 0) {
      filterResultItems.forEach((result, index) => {
        // Show the ones from the resultset:
        filterResultItems.forEach((item) => item.classList.remove('hidden--filter'))
      })
    }

    // Show results empty text if there are no visible results:
    let hasVisibleItems = false;
    _this._domGetItemElements().forEach((item, index) => {
      if ((item.offsetWidth > 0 || item.offsetHeight > 0)) {
        hasVisibleItems = true;
        return
      }
    });
    if (hasVisibleItems) {
      _this._domGetEmptyResultsElement().classList.add('hidden--results-empty');
    } else {
      _this._domGetEmptyResultsElement().classList.remove('hidden--results-empty');
    }
  }

  this._filterBySearch = function(filterValue){
    filterValue = filterValue.trim()
    if (filterValue.length == 0) {
      _this._domGetItemElements().forEach((item, index) => {
        item.classList.remove('hidden--search');
      })
      return;
    }
    // Split words by space (multi word or search):
    const words = filterValue.split(' ');
    _this._domGetItemElements().forEach((item, index) => {
      item.classList.remove('hidden--search');
      words.forEach((word) => {
        if(!item.textContent.toLowerCase().includes(word.toLowerCase())) {
          item.classList.add('hidden--search');
        }
      });
    });
  }

  this._filterByDataAttribute = function(dataAttribute, filterValues = []) {
    _this._domGetItemElements().forEach((item, index) => {

    });
  }

  this.reset = function() {
    // Reset search input:
    _this._domGetSearchInputElement().value = '';

    // Reset facet filters:
    _this._domGetFacetFilterInputElements().forEach((input, index) => {
      input.checked = false;
    })

    _this.filter(null);
  }

  this._domGetFacetsContainerElement = function(){
    return document.querySelector(settings._csff_controls_facets_container_selector);
  }

  this._domGetInjectionTargetElement = function(){
    return document.querySelector(settings.injection_selector);
  }

  this._domGetItemElements = function() {
    return targetElement.querySelectorAll(settings._csff_item_selector);
  }

  this._domGetSearchInputElement = function() {
    return this._domGetFacetsContainerElement().querySelector('.csff-search__value');
  }

  this._domGetFacetFilterInputElements = function() {
    return this._domGetFacetsContainerElement().querySelectorAll('.csff-facet-value__value');
  }

  this._domGetResetButtonInputElement = function() {
    return this._domGetFacetsContainerElement().querySelector('.csff-reset');
  }

  this._domGetEmptyResultsElement = function() {
    return targetElement.querySelector('.csff-results-empty');
  }

  this._theme_facets_container = function() {
    return _this._interpolateTemplate(settings.template_facets_container, {});
  }

  this._theme_facet = function(id, title, description, content) {
    return _this._interpolateTemplate(settings.template_facet, {
      id: id,
      title: title,
      description: description,
      content: content,
    })
  }

  this._theme_facet_value = function(facet_id, value, count = false) {
    const id = (facet_id + '-' + value).replace(/\W/g,'-').toLowerCase();
    return _this._interpolateTemplate(settings.template_facet_value, {
      facet_id: facet_id,
      value: value,
      id: id,
      count: count,
    })
  }

  this._theme_facet_reset = function (title = '', description = '', label = '') {
    const content = _this._interpolateTemplate(settings.template_facet_reset, {
      label: label,
    })
    const id = 'csff-reset'
    return _this._theme_facet(id, title, description, content);
  }

  this._theme_search = function (title = '', description = '', placeholder = '') {
    const content = _this._interpolateTemplate(settings.template_search, {
      placeholder: placeholder,
    })
    const id = 'csff-search'
    return _this._theme_facet(id, title, description, content);
  }

  this._theme_results_empty = function(content) {
    return _this._interpolateTemplate(settings.template_results_empty, {
      content: content,
    });
  }

  this._interpolateTemplate = function (templateString, replacements) {
    return Object.entries(replacements).reduce(
      (result, [key, value]) => result.replaceAll(`$\{${key}}`, `${value}`),
      templateString,
    )
  }

  // Call init and return self.
  this.init();
  return this;
}
